
#include <dos.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "flic.h"
#include "car.h"
#include "svga.h"
#include "lin.h"
#include "cgdk.h"
#include "timer.h"
#include "dpmi.h"

#define FLI_COLOR256 4   // 256-level color palette info
#define FLI_SS2 7        // Word-oriented delta compression
#define FLI_COLOR 11     // 64-level color palette info
#define FLI_LC 12        // Byte-oriented delta compression
#define FLI_BLACK 13     // Entire frame is color index
#define FLI_BRUN 15      // Byte run length compression
#define FLI_COPY 16      // No compression
#define FLI_PSTAMP 18    // Postage stamp sized image
#define FPS_BRUN 15      // Byte run length compression
#define FPS_COPY 16      // No compression
#define FPS_XLAT256 18   // Six-cube color xlate table

#define FLC_FILETYPE 0xAF12
#define FLI_FILETYPE 0xAF11

#define FLIC_FRAMECHUNK 0xF1FA
#define FLIC_LINESKIP 0xC000

int FLIC_framesCount;
int FLIC_length;
int FLIC_width;
int FLIC_delay;
long FLIC_bufSize;

typedef struct TFLICSubChunk
{
  long     size;
  unsigned short type;

} TFLIC_SubChunk;


typedef struct TFLICHeader       // FLIC header (128 bytes)
{
  long     fileSize;            // FLIC file size
  unsigned short fileType;            // file format identifier
  unsigned short frames;              // Number of frames in .FLIC file
  unsigned short screenLength;        //
  unsigned short screenWidth;        //
  unsigned short depth;               // Bites per pixel
  unsigned short flags;               //
  long     speed;               // Delay (ms) between each frame
  unsigned short reserved1;
  long     created;             // MS-DOS packed date and time
  char     creator [4];         // FLIB
  long     updated;             //
  char     updatetor [4];       //
  unsigned short aspectX;             // X ratio at which the file was created
  unsigned short aspectY;             // Y ratio, 320x200 ... ratio 6:5
  char     reserved2 [38];
  long     offrame1;            // offset from beginning file
                                // to 1st animation chunk
  long     offrame2;            // offset from beginning file
                                // to 2nd animation chunk (used for looping)
  char     reserved3 [40];
} TFLIC_header;


typedef struct TFLICChunk
{
  long     size;                // Chunk size inluding this header
                                // and followed sub-chunks
  unsigned short type;                // Frame chunk identifier
                                // F100 ... prefix chunk
                                // F1FA ... frame chunk
  unsigned short chunks;              // Number of followed sub-chunks
                                // if chunks = 0 then No Change
  char     reserved [8];
} TFLIC_chunk;



static void far *     FLIC_videoBuf;
int            FLIC_videoBufSeg;
int            FLIC_videoBufOff;
int            FLIC_videoBufSizeX;
int            FLIC_videoBufSizeY;


static TRGBColor *     FLIC_PALBuf;
static TFLIC_header   FLIC_header;
static long           FLIC_curOffset;
static int FLIC_frame2;
static int FLIC_curFrame;
static TCAR           *FLIC_carHandle;
static char           FLIC_fileName [80];
static TFLIC_chunk    FLIC_chunk;
static void *         FLIC_buf;

int FLIC_getInfo (TCAR *carHandle, const char *fileName, int *sizeX, int *sizeY)
{
  TFLIC_header flicHeader;

  if (CAR_seek (carHandle, fileName, 0) == -1)
    return 1;

  if (CAR_read (carHandle, &flicHeader, sizeof (TFLIC_header)) != sizeof (TFLIC_header))
    return 2;

  *sizeX = flicHeader.screenLength;
  *sizeY = flicHeader.screenWidth;

  return 0;
}

int FLIC_load (TCAR *carHandle, const char *fileName,
               void far *videoBuf, int videoBufSizeX, int videoBufSizeY,
               TRGBColor *PALBuf)

{
  FLIC_unload ();

  FLIC_videoBuf = videoBuf;
  FLIC_videoBufSeg = FP_SEG (videoBuf);
  FLIC_videoBufOff = FP_OFF (videoBuf);
  FLIC_videoBufSizeX = videoBufSizeX;
  FLIC_videoBufSizeY = videoBufSizeY;
  FLIC_PALBuf = PALBuf;

  FLIC_carHandle = carHandle;

  if (CAR_seek (FLIC_carHandle, fileName, 0) == -1)
    return 1;

  if ((FLIC_curOffset = CAR_read (FLIC_carHandle, &FLIC_header, sizeof (TFLIC_header))) != sizeof (TFLIC_header))
    return 2;

  if ((FLIC_header.screenLength > FLIC_videoBufSizeX) ||
      (FLIC_header.screenWidth > FLIC_videoBufSizeY))
    return 3;

  switch (FLIC_header. fileType)
  {
    case FLC_FILETYPE:
    {
      FLIC_delay = FLIC_header.speed;
      FLIC_curOffset = FLIC_header. offrame1;
      if (FLIC_header. offrame1 == FLIC_header. offrame2)
      {
        FLIC_frame2 = FLIC_curFrame;
      }
      break;
    }
    case FLI_FILETYPE:
    {
      FLIC_delay = (1000*FLIC_header.speed)/70;
      FLIC_header. offrame1 = FLIC_header. offrame2 = FLIC_curOffset;
      break;
    }
    default:
    {
    /* neznamy typ flic souboru */
      return 4;
    }
  }

  /* naalokovani bufferu pro zpracovani dat */
  FLIC_bufSize = (unsigned) (FLIC_header.screenLength * FLIC_header.screenWidth);
  FLIC_buf = _alloc (FLIC_bufSize);
  if (FLIC_buf == NULL)
    return 5;

  strcpy (FLIC_fileName, fileName);

  FLIC_curFrame = 0;
  FLIC_framesCount = FLIC_header.frames;
  FLIC_length = FLIC_header.screenLength;
  FLIC_width = FLIC_header.screenWidth;

  // read the first chunk
  CAR_seek (FLIC_carHandle, FLIC_fileName, FLIC_curOffset);
  CAR_read (FLIC_carHandle, &FLIC_chunk, sizeof (TFLIC_chunk));

  return 0;
}


static void FLIC_FLI_COLOR (unsigned char *flicPtr)
{
  TRGBColor *palPtr = FLIC_PALBuf;
  int i,j;
  int packetCount;
  int baseColor;
  int maxColor;

  packetCount = *(short int *)flicPtr;
  flicPtr += sizeof (short int);
  for (i = 0; i < packetCount; i++)
  {
    baseColor = *flicPtr;
    flicPtr ++;
    maxColor = *flicPtr;
    flicPtr++;
    if (maxColor == 0)
    {
      maxColor = 256;
    }

    for (j = 0; j < maxColor; j++, baseColor++)
    {
      palPtr [baseColor].r = *flicPtr * 4;
      flicPtr++;
      palPtr [baseColor].g = *flicPtr * 4;
      flicPtr++;
      palPtr [baseColor].b = *flicPtr * 4;
      flicPtr++;
      palPtr [baseColor].a = 0;
    }
  }
}


static void FLIC_FLI_COLOR256 (unsigned char *flicPtr)
{
  TRGBColor *palPtr = FLIC_PALBuf;
  int i,j;
  int packetCount;
  int baseColor;
  int maxColor;

  packetCount = *(short int *)flicPtr;
  flicPtr += sizeof (short int);
  for (i = 0; i < packetCount; i++)
  {
    baseColor = *flicPtr;
    flicPtr ++;
    maxColor = *flicPtr;
    flicPtr++;
    if (maxColor == 0)
    {
      maxColor = 256;
    }

    for (j = 0; j < maxColor; j++, baseColor++)
    {
      palPtr [baseColor].r = *flicPtr;
      flicPtr++;
      palPtr [baseColor].g = *flicPtr;
      flicPtr++;
      palPtr [baseColor].b = *flicPtr;
      flicPtr++;
      palPtr [baseColor].a = 0;
    }
  }
}

static void FLIC_COPY (unsigned char *flicPtr)
{
  int es;

  es = DPMI_getES ();

  DPMI_setES (FLIC_videoBufSeg);

  LIN_movWin_ ((void *)FLIC_videoBuf, FLIC_videoBufSizeX,
               flicPtr, FLIC_length, FLIC_length, FLIC_width);

  DPMI_setES (es);
}


#if defined __TURBOC__ || defined CGDK_TRIAL

static void FLIC_BRUN (unsigned char *flicPtr)
{
  int i,j;
  signed char c;
  char far *vbufPtr = (char far *) FLIC_videoBuf;

  for (i = 0; i < FLIC_width; i++)
  {
    flicPtr++;
    j = 0;
    while (j < FLIC_length)
    {
      c = *flicPtr;
      flicPtr++;

      if (c < 0)
      {
        c = -c;
        _fmemcpy (vbufPtr+j, flicPtr, c);
        flicPtr += c;
      }
      else
      {
        _fmemset (vbufPtr+j, *flicPtr, c);
        flicPtr++;
      }
      j += c;
    }
    vbufPtr += FLIC_videoBufSizeX;
  }
}

static void FLIC_LC (unsigned char *flicPtr)
{
  int i,j;
  int lines,columns;
  int skip;
  signed char c;
  char far *vbufPtr = (char far *) FLIC_videoBuf;

  vbufPtr += *(short int *) flicPtr*FLIC_videoBufSizeX;
  flicPtr += sizeof (short int);
  columns = *(short int *) flicPtr;
  flicPtr += sizeof (short int);

  for (i = 0; i < columns; i++)
  {
    lines = *flicPtr;
    flicPtr++;
    skip = 0;
    for (j = 0; j < lines; j++)
    {
      skip += *flicPtr;
      flicPtr++;

      c = *flicPtr;
      flicPtr++;
      if (c > 0)
      {
        _fmemcpy (vbufPtr+skip, flicPtr, c);
        flicPtr += c;
      }
      else
      {
        c = -c;
        _fmemset (vbufPtr+skip, *flicPtr, c);
        flicPtr ++;
      }
      skip += c;
    }
    vbufPtr += FLIC_videoBufSizeX;
  }
}

static void FLIC_SS2 (unsigned char *flicPtr)
{
  int i,j;
  int lines,columns;
  int skip;
  signed char c;
  char far *vbufPtr = (char far *) FLIC_videoBuf;

  lines = *(short int *) flicPtr;
  flicPtr += sizeof (short int);

  for (i = 0; i < lines; i++)
  {
    columns = *(short int *)flicPtr;
    flicPtr+=sizeof (short int);

    while ( ((unsigned short) columns & FLIC_LINESKIP) == FLIC_LINESKIP)
    {
      vbufPtr += -columns*FLIC_videoBufSizeX;
      columns = *(short int *)flicPtr;
      flicPtr+=sizeof (short int);
    }
/*    if (columns < 0)
    {
      columns = *(short int *)flicPtr;
      flicPtr+=sizeof (short int);
    }*/

    skip = 0;
    for (j = 0; j < columns; j++)
    {
      skip += *flicPtr;
      flicPtr++;
      c = *flicPtr;
      flicPtr++;
      if (c > 0)
      {
        _fmemcpy (vbufPtr+skip, flicPtr, c+c);
        flicPtr += c+c;
      }
      else
      {
        c = -c;
        _fmemset (vbufPtr+skip, *flicPtr, c+c);
        flicPtr += sizeof (short int);
      }
      skip += c+c;
    }
    vbufPtr += FLIC_videoBufSizeX;
  }
}


#else

extern "C" void cdecl FLIC_BRUN (unsigned char *flicBuf);

extern "C" void cdecl FLIC_LC (unsigned char *flicBuf);

extern "C" void cdecl FLIC_SS2 (unsigned char *flicBuf);

#endif




static void FLIC_proceedSubChunk (TFLIC_SubChunk *SubChunk)
{
  unsigned char *flicPtr = ((unsigned char *) SubChunk) + sizeof (TFLIC_SubChunk);

  switch (SubChunk->type)
  {
    case FLI_COLOR:
    {
      FLIC_FLI_COLOR (flicPtr);
      break;
    }

    case FLI_COLOR256:
    {
      FLIC_FLI_COLOR256 (flicPtr);
      break;
    }

    case FLI_BLACK:
    {
      memset (FLIC_PALBuf, 0, 256*sizeof (TRGBColor));
      break;
    }


    // prikazy pracujici s bitmapama

    case FLI_COPY:
    {
      FLIC_COPY (flicPtr);
      break;
    }

    case FLI_BRUN:
    {
      FLIC_BRUN (flicPtr);
      break;
    }

    case FLI_LC:
    {
      FLIC_LC (flicPtr);
      break;
    }

    case FLI_SS2:
    {
      FLIC_SS2 (flicPtr);
      break;
    }

    case FLI_PSTAMP:
    break;

  }
}



void FLIC_play (void)
{
  int i;
  char *subChunk;

  if (FLIC_buf != NULL)
  {
    subChunk = (char *) FLIC_buf;

    abs (1);

    /* vypocteni cisla druheho framu */
    if (FLIC_curOffset == FLIC_header. offrame2)
      FLIC_frame2 = FLIC_curFrame;

    /* test, zda nebyl zpracovan posledni frame */
    if (FLIC_curFrame > FLIC_header. frames)
    {
    /* pokud ano, nastaveni aktualni pozice na druhy frame */
      FLIC_curOffset = FLIC_header. offrame2;
      FLIC_curFrame = FLIC_frame2;
      CAR_seek (FLIC_carHandle, FLIC_fileName, FLIC_curOffset);
      CAR_read (FLIC_carHandle, &FLIC_chunk, sizeof (TFLIC_chunk));
    }
    else
    {
      /* vybrani aktualniho framu */
      CAR_seek (FLIC_carHandle, FLIC_fileName, FLIC_curOffset + sizeof (TFLIC_chunk));
    }

    /* vypocteni pozice dalsiho framu */
    FLIC_curOffset += FLIC_chunk.size;

    /* nacteni vsech sub-chunku */
    if (FLIC_chunk.size > FLIC_bufSize)
    {
      _free (FLIC_buf);
      FLIC_buf = _alloc (FLIC_chunk.size);
      if (FLIC_buf == NULL)
      {
        sprintf (exitStr, "FLIC buffer overflowed - requested %i bytes\n", FLIC_chunk.size);
        CGDK_exit (exitStr, EXIT_FAILURE);
      }
      FLIC_bufSize = FLIC_chunk.size;
    }

    CAR_read (FLIC_carHandle, FLIC_buf, FLIC_chunk.size);

    /* test zda byl nacten frame chunk */
    if (FLIC_chunk. type == FLIC_FRAMECHUNK)
    {
      if (FLIC_curFrame == 25)
      {
        abs (1);
      }
      /* cyklus pro nacteni vsech sub-chunku aktualniho chunku */
      for (i = 1; i <= FLIC_chunk. chunks; i++)
      {
        /* volani funkce pro zpracovani dat sub-chunku */
        FLIC_proceedSubChunk ( (TFLIC_SubChunk *) subChunk);

        /* ukazani na dalsi sub-chunk */
        subChunk += * ((long *) subChunk);
      }
      FLIC_curFrame ++;
    }

    memcpy ((void *)&FLIC_chunk, (char *)FLIC_buf+FLIC_chunk.size - sizeof (TFLIC_chunk), sizeof (TFLIC_chunk));
  }
}

void FLIC_unload (void)
{
  if (FLIC_buf != NULL)
  {
    _free (FLIC_buf);
    FLIC_buf = NULL;
  }
}

#include "flicp_.cpp"
